/* stereo_xquant.c      Heiko Purnhagen   21.04.93 */
/* modified for acc (double->float)       13.05.93 */
/* fixed e.g. reading of short files      18.05.93 */
/* fixed noise shaping                    19.05.93 */

/* see also: Psychoacoustically Optimal Noise Shaping               */
/* by:       Robert A. Wannemaker                                   */
/* in:       J. Audio Eng. Soc., Vol. 40, No. 7/8, 1992 July/August */

#include <stdio.h>
#include <stdlib.h>

#define MAXLEN 24000
#define STRLEN 256

short shortbuf[2*MAXLEN];
float fsample = 48000.;

int numread,numwrite;
int i,num,numsamples;
float starttime, lengthtime;
int startsample, lengthsample;

float fact_l,fact_r;
int quantsteps,min_value;
int final_shift,max_shift;
int dithering;
float noiseshape_fb;
int clipping,old_clipping;
int rv;

float error_l,error_r;
float tempf_l,tempf_r;
int tempi_l,tempi_r;

char infilename[STRLEN], outfilename[STRLEN];
FILE *infile;
FILE *outfile;


/* white noise with triangular pdf [-1 .. 1] */

float dithernoise (void)
{
  float temp;

  temp = (float) rand() / (float) RAND_MAX;
  temp += (float) rand() / (float) RAND_MAX;
  return (temp - 1.0);
}


main ()
{
  printf ("stereo-quant   Heiko Purnhagen   19.05.93\n");

  do {
    printf ("input filename (incl. .DAT) : ");
    scanf ("%s",infilename);
  } while ((infile = fopen(infilename,"rb")) == NULL);

  do {
    printf ("output filename (incl. .DAT) : ");
    scanf ("%s",outfilename);
  } while ((outfile = fopen(outfilename,"wb")) == NULL);

  printf ("f_sample = %f Hz\n",fsample);

  printf ("specify section to cut and quantize:\n");
  do {
    printf ("start of section (sec) : ");
    scanf ("%f",&starttime);
    startsample = (int) (starttime*fsample + 0.5);
  } while (startsample < 0);
  printf ("number of stereo samples to skip = %d\n",startsample);

  do {
    printf ("length of section (sec) : ");
    scanf ("%f",&lengthtime);
    lengthsample = (int) (lengthtime*fsample + 0.5);
  } while (lengthsample < 0);
  printf ("number of stereo samples to quantize = %d\n",lengthsample);

  do {
    printf ("amplification of left channel (factor) : ");
    scanf ("%f",&fact_l);
  } while (0);

  do {
    printf ("amplification of right channel (factor) : ");
    scanf ("%f",&fact_r);
  } while (0);

  do {
    printf ("number of quantizer steps (2 ..  65536) : ");
    scanf ("%d",&quantsteps);
  } while (quantsteps < 2 || quantsteps > 65536);
  min_value = - (quantsteps / 2);
  printf ("quantized values in [%d .. %d]\n",min_value,min_value+quantsteps-1);

  i = quantsteps;
  max_shift = 0;
  while (i <= 32768) {
    i *= 2;
    max_shift++;
  }
  do {
    printf ("final shift (left) (0 .. %d bits) : ",max_shift);
    scanf ("%d",&final_shift);
  } while (final_shift < 0 || final_shift > max_shift);
  if ((quantsteps/2)*2 == quantsteps && final_shift > 0)
    rv = 1 << (final_shift-1);
  else
    rv = 0;

  do {
    printf ("dithering (0=off(0dB) 1=on(+4.77dB)) : ");
    scanf ("%d",&dithering);
  } while (dithering < 0 || dithering > 1);

  do {
    printf ("noiseshape_feedback (factor)\n");
    printf ("(0.0 = off(0dB)  0.9 = -5.6dB..20dB  1.0 = -6.0dB..instable) : ");
    scanf ("%f",&noiseshape_fb);
  } while (noiseshape_fb < -1.0 || noiseshape_fb > 1.0);

  printf ("seeking\n");
  i = startsample;
  while (i > 0) {
    printf (".");
    i -= MAXLEN;
  }
  printf (" \n");

  while (startsample > 0) {
    num = (startsample > MAXLEN) ? MAXLEN : startsample;

    printf ("r");
    numread = fread (shortbuf,sizeof(short),2*num,infile);
    if (numread < 2*num) {
      printf ("error: file to short\n");
      return (-1);
    } 

    fflush (stdout);
    startsample -= num;
  }
  printf (" \n");

  printf ("quantizing\n");
  i = lengthsample;
  while (i > 0) {
    printf ("....");
    i -= MAXLEN;
  }
  printf (" \n");

  error_l = error_r = 0.0;
  old_clipping = clipping = 0;
  numsamples = 0;
  while (lengthsample > 0) {
    num = (lengthsample > MAXLEN) ? MAXLEN : lengthsample;

    printf ("r");
    numread = fread (shortbuf,sizeof(short),2*num,infile);
    if (numread < 2*num) {
      printf (" \n");
      printf ("error: input file to short\n");
      num = numread / 2;
      lengthsample = 0;
    }

    printf ("q");
    for (i=0; i<num; i++) {
      tempf_l = shortbuf[2*i] / 65536.0 * fact_l + 0.5;
      tempf_r = shortbuf[2*i+1] / 65536.0 * fact_r + 0.5;
      tempf_l = (tempf_l * quantsteps) - error_l * noiseshape_fb;
      tempf_r = (tempf_r * quantsteps) - error_r * noiseshape_fb;
      tempi_l = (int) (tempf_l + dithernoise() * dithering + 1) - 1;
      tempi_r = (int) (tempf_r + dithernoise() * dithering + 1) - 1;
      if (tempi_l < 0) {
	tempi_l = 0;
	clipping++;
      }
      if (tempi_l > quantsteps-1) {
	tempi_l = quantsteps-1;
	clipping++;
      }
      if (tempi_r < 0) {
	tempi_r = 0;
	clipping++;
      }
      if (tempi_r > quantsteps-1) {
	tempi_r = quantsteps-1;
	clipping++;
      }
      error_l = (float) tempi_l - tempf_l + 0.5;
      error_r = (float) tempi_r - tempf_r + 0.5;
      shortbuf[2*i] = (short) (((tempi_l + min_value) << final_shift) + rv);
      shortbuf[2*i+1] = (short) (((tempi_r + min_value) << final_shift) + rv);
    }
    if (clipping == old_clipping)
      printf ("-");
    else
      printf ("c");
    old_clipping = clipping;

    printf ("w");
    numwrite = fwrite (shortbuf,sizeof(short),2*num,outfile);
    if (numwrite < 2*num) {
      printf ("error: can't write output file\n");
      return (-1);
    }

    fflush (stdout);
    lengthsample -= num;
    numsamples += num;
  }
  printf (" \n");

  printf ("%d stereo samples quantized and written (%7.3f sec)\n",
	  numsamples,numsamples/fsample);
  printf ("%d samples clipped\n",clipping);
  fclose (outfile);
  fclose (infile);
  exit (0);
}
